## -*-Tcl-*-
 # ###################################################################
 #  Alphatk - the ultimate editor
 # 
 #  FILE: "alpha_positions.tcl"
 #                                    created: 04/12/98 {23:17:46 PM} 
 #                                last update: 04/23/2002 {05:53:09 PM} 
 #  Author: Vince Darley
 #  E-mail: vince.darley@kagi.com
 #    mail: Flat 10, 98 Gloucester Terrace, London W2 6HP
 #     www: http://www.santafe.edu/~vince/Alphatk.html
 #  
 # Copyright (c) 1998-2002  Vince Darley
 # 
 # See the file "license.terms" for information on usage and redistribution
 # of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 # 
 # In particular, while this is 'open source', it is NOT free, and
 # cannot be copied in full or in part except according to the terms
 # of the license agreement.
 # 
 # ###################################################################
 ##

namespace eval pos {}

# pos::toRowCol <pos> - converts from index position to row, column.
#  Here 'column' counts multiple times for each tab character and is
#  therefore a visual reference and may depend on the current window's
#  tabSize.
proc pos::toRowCol {args} {
    if {[lindex $args 0] == "-w"} {
	set w   [lindex $args 1]
	set pos [lindex $args 2]
    } else {
	set w   [win::Current]
	set pos [lindex $args 0]
    }
    set row [lindex [split [text_wcmd $w index $pos] .] 0]
    set line [getText -w $w [::lineStart -w $w $pos] $pos]
    set line [text::maxSpaceForm -w $w $line]
    list $row [string length $line]
}
# pos::toRowChar <pos> - converts from index position to row, char.
#  Here 'char' counts just once for each 'tab' character.
proc pos::toRowChar {args} {
    if {[lindex $args 0] == "-w"} {
	set p [text_wcmd $w index [lindex $args 2]]
    } else {
	set p [text_cmd index [lindex $args 0]]
    }
    return [split $p .]
}
# pos::fromRowCol [-w <win>] <row> <col> - converts to window position.
#  Accepts optional -w parameter that allows window to be specified.
#  Not sure if this is ok, but can't be too bad!
proc pos::fromRowCol {row col args} {
    if {$row == "-w"} {
	set w $col
	foreach {row col} $args {break}
    } else {
	set w [win::Current]
    }
    if {$col < 0} {
	return -code error "Negative column!"
    } elseif {$col == 0} {
	return ${row}.0
    }
    set lineStart "${row}.0"
    set lineEnd [lindex [split [text_wcmd $w index "${row}.0 lineend"] .] 1]
    set char 0
    while {$char < $lineEnd} {
	set line [getText -w $w $lineStart ${row}.$char]
	set len [string length [text::maxSpaceForm -w $w $line]]
	if {$len > $col} {
	    break
	}
	incr char
    }
    
    return $row.$char
}
# pos::fromRowChar [-w <win>] <row> <char> - converts to window position.
#  Accepts optional -w parameter that allows window to be specified.
proc pos::fromRowChar {row col args} {
    if {$row == "-w"} {
	return "[lindex $args 0].[lindex $args 1]"
    } else {
	return $row.$col
    }
}

