## -*-Tcl-*-
 # ==========================================================================
 #  WWW Menu - an extension package for Alpha
 # 
 #  FILE: "wwwMode.tcl"
 #                                    created: 04/30/1997 {11:04:46 am} 
 #                                last update: 05/20/2002 {12:45:44 PM} 
 #  Author: Vince Darley
 #  E-mail: <vince@santafe.edu>
 #    mail: 317 Paseo de Peralta, Santa Fe, NM 87501, USA
 #     www: <http://www.santafe.edu/~vince/>
 #     
 #  Description:
 #  
 #  The WWW Menu provides a simple text based (Lynx-like) HTML file
 #  browser, for reading local HTML files
 # 
 #  The reason to create a "WWW" mode is so that rendered windows can
 #  have easy access to the menu, and appropriate keybindings.
 # 
 #  See the "wwwVersionHistory.tcl" file for license info, credits, etc.
 # ==========================================================================
 ##

alpha::mode WWW {for wwwMenu menu} {wwwMode.tcl} {} {
    wwwMenu
} {
    # This requires AlphaTcl 7.6d1 for new position procs.
    alpha::package require -loose AlphaTcl 7.6d1
    set "newDocTypes(New WWW Browser)" WWW::renderUrl
    set "htmlViewer(Text-only parser)" WWW::renderFile
    if {[info tclversion] >= 8.0} {
	set "urlViewer(Internal text-only viewer)" WWW::renderUrl
    }
}

proc wwwMode.tcl {} {}

namespace eval WWW {}

# As of this writing, Alpha7 cannot handle the downloading of remote urls, so
# we disable that 'View Url' and associated menu items. The condition
# 
# {    } elseif {$tcl_platform(platform) == "macintosh"} {...}
# 
# really shouldn't be here, but as of this writing 'httpCopy' will do a nasty
# freeze in Alpha8.0b1, so we only offer this for Alphatk and AlphaX.
# 
# To debug this, just make this proc return 1, evaluate this proc, and then
# use the AlphaDev menu item 'Rebuild A Menu' to rebuild the WWW Menu.  This
# will enable the 'View Url' and related menu items and support procs.

proc WWW::httpAllowed {} {
    
    global tcl_platform
    
    if {[info tclversion] < 8.0} {
	return 0
    } elseif {$tcl_platform(platform) == "macintosh"} {
        return 0
    } elseif {![catch {package present http}]} {
        return 1
    } elseif {![catch {package require http}]} {
	return 1
    } else {
        return 0
    }
}

# ===========================================================================
#
#  WWW prefs, vars, arrays  #
# 

# We let the global pref for 'Compose Email Using ...' deal with this now.
prefs::removeObsolete WWWmodeVars(mailtoLinksInternal)

# To automatically adjust the text in a rendered window to fit its width,
# turn this item on||To always use the value of the WWW mode preference
# 'Fill Column' when rendering windows, turn this item off
newPref flag autoAdjustWidth        1 WWW
# To automatically 'float' a palette containing a rendered window's marks,
# turn this item on|| after To disable the automatic 'floating' of a palette
# containing a rendered window's marks, turn this item off
newPref flag autoFloatMarks         0 WWW
# To only open one WWW browser window at a time, turn this item on. 
# (Navigating bookmarks or history items will kill the current window)||To
# create a new window for every WWW browser page, turn this item off
newPref flag linksOpenNewWindow     1 WWW
# To automatically place the selected link in the center of the window when
# navigating up or down, turn this item on||To never automatically place
# the selected link in the center of the window when navigating up or down,
# turn this item on 
newPref flag centerRefreshOnNav     0 WWW
# To handle all 'ftp' links internally, turn this item
# on||To never handle any 'ftp' links internally, turn this item off
newPref flag ftpLinksInternal       0 WWW
# To fetch and parse all 'http' links using the WWW menu, turn this item
# on||To never send any fetch and parse all 'http' links using the WWW
# menu, turn this item off
newPref flag httpLinksInternal      1 WWW
# To ignore all forms while rendering, turn this item on||To insert 'form
# begins/end' markers when rendering forms, turn this item off
newPref flag ignoreForms            0 WWW
# To ignore all images while rendering, turn this item on||To insert 'image'
# markers when rendering files, turn this item off
newPref flag ignoreImages           0 WWW
# To send all 'java applets' links using the 'Java Viewer Sig', turn this
# item on||To never send any 'mailto' links to Alpha's Eudora menu, turn
# this item off
newPref flag runJavaAppletsDirectly 0 WWW
# To send all unrecognized links (or those turned off by the prefs above) to
# internet config, turn this item on||To never send any unrecognized link to
# internet config, turn this item off
newPref flag wwwSendRemoteLinks     0 WWW

# This local file or remote url will be opened by the "Home Page" menu item.\
newPref url homePage "" WWW

if {[info tclversion] < 8.0} {
    newPref url homePage [file::toUrl [file join $HOME Help "HTML Help" HTMLmanual.html]] WWW
} else {
    newPref url homePage "http://alphatcl.sourceforge.net/wikit/" WWW
    if {$tcl_platform(platform) != "unix"} {
	# Default font to use for new WWW windows
	newPref var wwwFont     $defaultFont WWW {WWW::refreshFont} "system monaco courier geneva helvetica profont times {new york} programmer"
	# Default font size to use for new WWW windows
	newPref var wwwFontSize $fontSize    WWW {WWW::refreshFont} "7 9 10 12 14 18"
    } else {
	# Default font to use for new WWW windows
	newPref var wwwFont     $defaultFont WWW {WWW::refreshFont} "system fixed monaco courier geneva helvetica profont times {new york} programmer"
	# Default font size to use for new WWW windows
	newPref var wwwFontSize $fontSize    WWW {WWW::refreshFont} "7 9 10 12 14 18"
    }
    # This is called when the user changes the font, fontsize.
    proc WWW::refreshFont {args} {
	global mode wwwMenu
	if {$mode == "WWW"} {WWW::menuProc $wwwMenu "refresh"}
    }
}

# This is the maximum width used before wrapping text when rendering windows.
newPref var fillColumn             75 WWW

newPref sig javaViewerSig "WARZ"

if {[set alpha::platform] == "tk"} {
    # In Alphatk, shadow, outline, and bold all turn text blue, so they
    # aren't very useful as options here.
    set wwwHeaderStyles [list underline - normal]
    set wwwHeader1      "underline"
    set wwwHeader2      "underline"
    set wwwHeader3      "underline"
} else {
    set wwwHeaderStyles [list shadow outline bold underline italics - normal]
    set wwwHeader1      "outline"
    set wwwHeader2      "bold"
    set wwwHeader3      "underline"
}

newPref var header1Style $wwwHeader1 WWW "" $wwwHeaderStyles
newPref var header2Style $wwwHeader2 WWW "" $wwwHeaderStyles
newPref var header3Style $wwwHeader3 WWW "" $wwwHeaderStyles

unset wwwHeaderStyles wwwHeader1 wwwHeader2 wwwHeader3

newPref color header1Color     blue     WWW
newPref color header2Color     red      WWW
newPref color header3Color     red      WWW
newPref color linkColor        green    WWW
newPref color visitedLinkColor magenta  WWW

# Command double-clicking will send the highlighted text to this search
# engine.
newPref url searchUrl1 {http://www.google.com/search?q=} WWW
# Command double-clicking while pressing the "option" key will send the
# highlighted text to this search engine.
newPref url searchUrl2 {http://search.metacrawler.com/crawler?general=} WWW
# Command double-clicking while pressing the "control" key will send the
# highlighted text to this search engine.
newPref url searchUrl3 {http://www.altavista.com/sites/search/web?q=} WWW
# Command double-clicking while pressing the "shift" key will send the
# highlighted text to this search engine.
newPref url searchUrl4 {http://google.yahoo.com/bin/query?p=} WWW

# Alpha 7 can't handle these yet.
if {![WWW::httpAllowed]} {
    unset WWWmodeVars(httpLinksInternal)
    unset WWWmodeVars(searchUrl1)
    unset WWWmodeVars(searchUrl2)
    unset WWWmodeVars(searchUrl3)
    unset WWWmodeVars(searchUrl4)
} 

# Contextual Menu modules
if {[info tclversion] > 8.0} {
    newPref flag bookmarksMenu   1 contextualMenuWWW
    newPref flag goToPageMenu    1 contextualMenuWWW
    newPref flag wwwLinksMenu    1 contextualMenuWWW
    newPref flag wwwWindowMenu   1 contextualMenuWWW
    newPref flag pageForwardItem 1 contextualMenuWWW
    newPref flag pageBackItem    1 contextualMenuWWW
    hook::register contextualPostBuildHook {WWW::postEvalCM} "WWW"
} 

# These prefs will appear in the menu.

set WWW::PrefsInMenu {
    linksOpenNewWindow
    centerRefreshOnNav
    autoAdjustWidth
    autoFloatMarks
    ignoreForms
    ignoreImages
    (-)
}

foreach pref [lsort -ignore [array names WWWmodeVars]] {
    if {[regexp {Links|Applets} $pref]} {
	lappend WWW::PrefsInMenu $pref
    } 
}
unset pref

# Links, history vars

ensureset WWW::Pages        ""
ensureset WWW::PagePos      -1
ensureset WWW::Visited      ""
ensureset WWW::FileSource() [list ]
ensureset WWW::BaseUrl()    [list ]

ensureset WWW::History [list [mtime [now] short]]
prefs::modified WWW::History

# To perform a special action with a new URL type, add an array entry
# indicating the procedure to be called with the regexp of the URL. Patterns
# which do NOT start with '^' are checked first, see 'WWW::link' for more
# information.  Packages which define url handlers will not have their
# regexps over-written here, so long as they've been defined before this file
# is sourced.  It is up to individual url handlers to decide if the action
# should be dealt with internally or shipped off to internet config, possibly
# with a user preference for the action (such as those defined above.)

ensureset WWW::UrlAction(^mailto:)   "composeEmail"
ensureset WWW::UrlAction(^ftp://)    "WWW::ftpLink"
ensureset WWW::UrlAction(^file://)   "WWW::fileLink"
ensureset WWW::UrlAction(^http://)   "WWW::httpLink"

ensureset WWW::UrlAction(\.class\$)  "WWW::javaLink"
ensureset WWW::UrlAction(\.gif\$)    "WWW::externalLink"
ensureset WWW::UrlAction(\.java\$)   "WWW::javaLink"

# We know that we can't handle any of these files internally.

foreach suffix [list aiff au gif jpe jpeg jpg mov mpeg mpg \
  pdf pict ps tif tiff] {
    ensureset WWW::UrlAction(\.$suffix\$) "WWW::externalLink"
}

unset suffix

# This is how we can completely redirect some tags to procs, concievably
# bypassing 'WWW::link' altogether.

ensureset WWW::UrlAction(WWW::indexSearch) "WWW::indexSearch"

# These are the parameters used for new WWW browser windows.  See the
# "wwwWindows.tcl" file for more information.

ensureset WWW::WindowParameters(Default) \
  [list $defLeft $defTop $defWidth $defHeight]
prefs::modified WWW::WindowParameters(Default)

# ===========================================================================
#
#  Html Style, Accent Elements  #
# 

# We are going to render 'CODE' 'KBD' 'SAMP' and 'TT' in plain text, so we
# don't define them here, but actually remove the tags in 'WWW::removeCrap'. 
# Since Alphatk doesn't do anything with bold etc, we simplify this to make
# it more obvious what's going to happen.

if {[set alpha::platform] == "alpha"} {
    
    array set WWW::HtmlToStyle {
	"B"         bold
	"BIG"       outline
	"CITE"      italic
	"DFN"       italic
	"EM"        italic
	"I"         italic
	"SMALL"     condensed
	"STRONG"    bold
	"U"         underline
	"VAR"       italic
    }

} else {
    
    array set WWW::HtmlToStyle {
	"B"         blue
	"BIG"       blue
	"CITE"      italic
	"DFN"       blue
	"EM"        blue
	"I"         blue
	"SMALL"     blue
	"STRONG"    blue
	"U"         underline
	"VAR"       blue
    }

}

array set WWW::HtmlToAccents {

    "#9"        {}
    "#10"       "\r"
    "#13"       "\r"
    "#32"       { }
    "#33"       {!}
    "#34"       {\"}
    "#35"       {\#}
    "#36"       {$}
    "#37"       {%}
    "#38"       {&}
    "#39"       {'}

    "#40"       {\(}
    "#41"       {\)}
    "#42"       {*}
    "#43"       {+}
    "#44"       {,}
    "#45"       {-}
    "#46"       {.}
    "#47"       {/}
    "#48"       {0}
    "#49"       {1}

    "#50"       {2}
    "#51"       {3}
    "#52"       {4}
    "#53"       {5}
    "#54"       {6}
    "#55"       {7}
    "#56"       {8}
    "#57"       {9}
    "#58"       {:}
    "#59"       {;}

    "#60"       {<}
    "#61"       {=}
    "#62"       {>}
    "#63"       {?}
    "#64"       {@}
    "#65"       {A}
    "#66"       {B}
    "#67"       {C}
    "#68"       {D}
    "#69"       {E}

    "#70"       {F}
    "#71"       {G}
    "#72"       {H}
    "#73"       {I}
    "#74"       {J}
    "#75"       {K}
    "#76"       {L}
    "#77"       {M}
    "#78"       {N}
    "#79"       {O}

    "#80"       {P}
    "#81"       {Q}
    "#82"       {R}
    "#83"       {S}
    "#84"       {T}
    "#85"       {U}
    "#86"       {V}
    "#87"       {W}
    "#88"       {X}
    "#89"       {Y}

    "#90"       {Z}
    "#91"       {\[}
    "#92"       {\\}
    "#93"       {\]}
    "#94"       {^}
    "#95"       {_}
    "#96"       {`}
    "#97"       {a}
    "#98"       {b}
    "#99"       {c}

    "#100"      {d}
    "#101"      {e}
    "#102"      {f}
    "#103"      {g}
    "#104"      {h}
    "#105"      {i}
    "#106"      {j}
    "#107"      {k}
    "#108"      {l}
    "#109"      {m}

    "#110"      {n}
    "#111"      {o}
    "#112"      {p}
    "#113"      {q}
    "#114"      {r}
    "#115"      {s}
    "#116"      {t}
    "#117"      {u}
    "#118"      {v}
    "#119"      {w}

    "#120"      {x}
    "#121"      {y}
    "#122"      {z}
    "#123"      {{}
    "#124"      {|}
    "#125"      {}}
    "#126"      {~}
    "#127"      {}

    "#130"      {}
    "#131"      {}
    "#132"      {}
    "#133"      {}
    "#134"      {}
    "#135"      {}
    "#136"      {}
    "#137"      {}
    "#138"      {?}
    "#139"      {}

    "#145"      {} 
    "#146"      {} 
    "#147"      {} 
    "#148"      {} 
    "#149"      {} 

    "#150"      {} 
    "#151"      {} 
    "#152"      {} 
    "#153"      {} 
    "#154"      {?} 
    "#155"      {} 
    "#156"      {} 
    "#159"      {}

    "#160"	{ }
    "#161"	{}
    "#162"	{}
    "#163"	{}
    "#164"	{?}
    "#165"	{}
    "#166"	{|}
    "#167"	{}
    "#168"	{}
    "#169"      {}

    "#170"	{}
    "#171"	{}
    "#172"	{}
    "#175"	{}
    "#176"	{}
    "#177"	{}

    "#180"	{}
    "#181"	{}
    "#182"	{}
    "#183"	{}
    "#184"	{}
    "#186"	{}
    "#187"	{}

    "#191"	{}
    "#247"	{}
    "#338"	{}
    "#339"	{}
    "#376"	{}
    "#402"	{}

    "#8211"	{}
    "#8212"	{}
    "#8216"	{}
    "#8217"	{}
    "#8218"	{}
    "#8220"	{}
    "#8221"	{}
    "#8222"	{}
    "#8224"	{}
    "#8225"	{}
    "#8226"	{}
    "#8230"	{}
    "#8240"	{}
    "#8249"	{}
    "#8250"	{}
    "#8482"	{}
    "#8730"	{}
    "#8734"	{}
    "#8747"	{}
    "#8776"	{}
    "#8800"	{}
    "#8804"	{}
    "#8805"	{}

    "aacute"	{}
    "Aacute"	{}
    "acirc"	{}
    "Acirc"	{}
    "aelig"	{}
    "AElig"	{}
    "agrave"	{}
    "Agrave"	{}
    "aring"	{}
    "Aring"	{}
    "atilde"	{}
    "Atilde"	{}
    "auml"	{}
    "Auml"	{}
    "ccedil"	{}
    "Ccedil"	{}
    "copy"	{}
    "eacute"	{}
    "Eacute"	{}
    "ecirc"	{}
    "Ecirc"	{}
    "Euml"	{}
    "egrave"	{}
    "Egrave"	{}
    "euml"	{}
    "gt"	{>}
    "iacute"	{}
    "Iacute"	{}
    "icirc"	{}
    "Icirc"	{}
    "igrave"	{}
    "Igrave"	{}
    "iuml"	{}
    "Iuml"	{}
    "lt"	{<}
    "nbsp"	{ }
    "ntilde"	{}
    "Ntilde"	{}
    "oacute"	{}
    "Oacute"	{}
    "ocirc"	{}
    "Ocirc"	{}
    "ograve"	{}
    "Ograve"	{}
    "oslash"	{}
    "Oslash"	{}
    "otilde"	{}
    "Otilde"	{}
    "ouml"	{}
    "Ouml"	{}
    "quot"	{"}
    "reg"	{}
    "szlig"	{}
    "uacute"	{}
    "Uacute"	{}
    "ucirc"	{}
    "Ucirc"	{}
    "ugrave"	{}
    "Ugrave"	{}
    "uuml"	{}
    "Uuml"	{}
    "yuml"	{}
}

# ===========================================================================
# 
#  Bindings  #
# 
# Bind various keys to imitate lynx.  Please route everything through the
# menu procs to ensure consistent handling.
# 
#  +++ Keystroke Commands	+++
# 
#  MOVEMENT:  Down arrow	- Highlight next topic
#             Up arrow	        - Highlight previous topic
#             Right arrow	- Jump to highlighted topic
#             Return, Enter	- Jump to highlighted topic
#             Left arrow	- Return to previous topic
# 
#  SCROLLING: +                 - Scroll down to next page (Page-Down)
#             -                 - Scroll up to previous page (Page-Up)
#             SPACE             - Scroll down to next page (Page-Down)
#             b                 - Scroll up to previous page (Page-Up)
#             CTRL-A            - Go to first page of the current document (Home)
#             CTRL-E            - Go to last page of the current document (End)
#             CTRL-B            - Scroll up to previous page (Page-Up)
#             CTRL-F            - Scroll down to next page (Page-Down)
#             CTRL-N            - Go forward two lines in the current document
#             CTRL-P            - Go back two lines in the current document
#             )                 - Go forward half a page in the current document
#             (                 - Go back half a page in the current document

Bind  'a'       {WWW::navigationProc bookmarks addBookmark} WWW
Bind  'b'       {WWW::navigationProc goToPage back}         WWW
Bind  'c'       {WWW::menuProc "" copyLinkLocation}         WWW
Bind  'e'       {WWW::menuProc "" editLinkedDocument}       WWW
Bind  'f'       {WWW::navigationProc goToPage forward}      WWW
Bind  'g'       {WWW::navigationProc goToPage ""}           WWW
Bind  'r'       {WWW::menuProc "" reload}                   WWW
Bind  'r' <c>   {WWW::menuProc "" refresh}                  WWW
Bind  'h'       {WWW::menuProc "" history}                  WWW
Bind  'h' <c>   {WWW::menuProc "" home}                     WWW
Bind  'o'       {WWW::menuProc "" openSourceInAlpha}        WWW
Bind  's'       {WWW::menuProc "" sendSourceToBrowser}      WWW
Bind  's' <c>   {WWW::menuProc "" saveSourceFileAs}         WWW
Bind  'v'       {WWW::renderUrl}                            WWW

Bind 0x73 <c>   {WWW::menuProc "" home}                     WWW
Bind 0x24 <z>   {WWW::menuProc "" modifyLink}               WWW
Bind 0x24 <o>   {WWW::menuProc "" editLinkedDocument}       WWW
Bind 0x24 <s>   {WWW::menuProc "" linkInNewWindow}          WWW
Bind 0x24 <zs>  {WWW::menuProc "" linkToBrowser}            WWW

Bind Home       {WWW::navigationProc "" Home}               WWW
Bind  'a' <z>   {WWW::navigationProc "" Home}               WWW
Bind End        {WWW::navigationProc "" End}                WWW
Bind  'e' <z>   {WWW::navigationProc "" End}                WWW

# Page up and page down should just scroll the window!
#Bind 0x79       {WWW::navigationProc "" PageForward}        WWW
#Bind 0x74       {WWW::navigationProc "" PageBack}           WWW
Bind 0x31       {WWW::navigationProc "" PageForward}        WWW
Bind  '+'       {WWW::navigationProc "" PageForward}        WWW
Bind  '-'       {WWW::navigationProc "" PageBack}           WWW

Bind  'n' <z>   {WWW::navigationProc "" twoLinesForward}    WWW
Bind  'p' <z>   {WWW::navigationProc "" twoLinesBack}       WWW
Bind  ')'       {WWW::navigationProc "" halfPageForward}    WWW
Bind  '('       {WWW::navigationProc "" halfPageBack}       WWW

Bind 0x7b       {WWW::navigationProc goToPage back}         WWW
Bind '\[' <c>   {WWW::navigationProc goToPage back}         WWW

Bind 0x7c       {WWW::link}                                 WWW
Bind '\]' <c>   {WWW::navigationProc goToPage forward}      WWW

Bind  '\t'      {WWW::navigationProc "" down}               WWW
Bind 0x7d       {WWW::navigationProc "" down}               WWW

Bind 0x7e       {WWW::navigationProc "" up}                 WWW

Bind 0x24       {WWW::link}                                 WWW
Bind 0x34       {WWW::link}                                 WWW

#  More mode stuff  #

proc WWW::DblClick {from to shift option control} {
    
    if {![WWW::httpAllowed]} {
        status::msg "Command double click is not yet implemented."
	return
    } 
    
    global WWWmodeVars
    
    if {![catch {WWW::getCurrentLink} result]} {
	WWW::link [lindex $result 3]
    } else {
	select $from $to
	set text [getSelect]
	# Any modifiers pressed?
	if {$option} {
	    WWW::searchProc "" wwwSearch2 $text
	} elseif {$control} {
	    WWW::searchProc "" wwwSearch3 $text
	} elseif {$shift} {
	    WWW::searchProc "" wwwSearch4 $text
	} else {
	    WWW::searchProc "" wwwSearch1 $text
	} 
    }
}

# Marks should be set be whatever proc is creating a WWW window, and
# then store them in a WWW::Marks cache.  This is simply to avoid the
# annoyance of wiping out the set of marks previously set.

proc WWW::MarkFile {} {
    
    global WWW::Marks
    
    if {![info exists WWW::Marks([set title [win::Current]])]} {
	status::msg "No marks found in '$title'."
	return
    }
    foreach mark [set marks [set WWW::Marks($title)]] {
	eval setNamedMark $mark
    }
    if {![llength $marks]} {
	status::msg "'$title' doesn't contains any marks."
    } elseif {[llength $marks] == "1"} {
	status::msg "'$title' contains $count mark."
    } else {
	status::msg "'$title' contains $count marks."
    }
}

proc WWW::parseFuncs {} {

    foreach link [WWW::getLinks] {
	set pos0 [lindex $link 0] 
	set pos1 [lindex $link 1] 
	set name [win::MakeTitle [getText $pos0 $pos1]]
	lappend results [list $name $pos0]
    }
    return [join [lsort -ignore $results]]
}

# ===========================================================================
# 
# .